@@ -3,7 +3,7 @@ extern crate polytype;
 extern crate programinduction;
 extern crate rand;
 use programinduction::pcfg::{self, Grammar, Rule};
-use programinduction::{GPParams, Task, GP};
+use programinduction::{GPParams, GPSelection, Task, GP};
 use rand::{rngs::SmallRng, SeedableRng};
 
 #[test]
@@ -38,6 +38,7 @@ fn gp_sum_arith() {
     };
 
     let gpparams = GPParams {
+        selection: GPSelection::Deterministic,
         population_size: 10,
         tournament_size: 5,
         mutation_prob: 0.6,
